"""AEMET OpenData Station."""

from datetime import datetime
from typing import Any
from zoneinfo import ZoneInfo

from .const import (
    AEMET_ATTR_IDEMA,
    AEMET_ATTR_STATION_ALTITUDE,
    AEMET_ATTR_STATION_DATE,
    AEMET_ATTR_STATION_DEWPOINT,
    AEMET_ATTR_STATION_HUMIDITY,
    AEMET_ATTR_STATION_LATITUDE,
    AEMET_ATTR_STATION_LOCATION,
    AEMET_ATTR_STATION_LONGITUDE,
    AEMET_ATTR_STATION_PRECIPITATION,
    AEMET_ATTR_STATION_PRESSURE,
    AEMET_ATTR_STATION_PRESSURE_SEA,
    AEMET_ATTR_STATION_TEMPERATURE,
    AEMET_ATTR_STATION_TEMPERATURE_MAX,
    AEMET_ATTR_STATION_TEMPERATURE_MIN,
    AEMET_ATTR_STATION_WIND_DIRECTION,
    AEMET_ATTR_STATION_WIND_SPEED,
    AEMET_ATTR_STATION_WIND_SPEED_MAX,
    AOD_ALTITUDE,
    AOD_COORDS,
    AOD_DATETIME,
    AOD_DEW_POINT,
    AOD_DISTANCE,
    AOD_HUMIDITY,
    AOD_ID,
    AOD_NAME,
    AOD_OUTDATED,
    AOD_PRECIPITATION,
    AOD_PRESSURE,
    AOD_TEMP,
    AOD_TEMP_MAX,
    AOD_TEMP_MIN,
    AOD_TIMESTAMP,
    AOD_TIMEZONE,
    AOD_WIND_DIRECTION,
    AOD_WIND_SPEED,
    AOD_WIND_SPEED_MAX,
    ATTR_DATA,
    ATTR_DISTANCE,
    STATION_MAX_DELTA,
)
from .helpers import get_current_datetime, parse_api_timestamp, timezone_from_coords


class Station:
    """AEMET OpenData Station."""

    altitude: float
    coords: tuple[float, float]
    _datetime: datetime
    distance: float
    dew_point: float
    humidity: float
    id: str
    name: str
    precipitation: float
    pressure: float
    temp: float
    temp_max: float
    temp_min: float
    wind_direction: float | None
    wind_speed: float | None
    wind_speed_max: float | None
    zoneinfo: ZoneInfo

    def __init__(self, data: dict[str, Any]) -> None:
        """Init AEMET OpenData Station."""
        self.altitude = float(data[AEMET_ATTR_STATION_ALTITUDE])
        self.coords = (
            float(data[AEMET_ATTR_STATION_LATITUDE]),
            float(data[AEMET_ATTR_STATION_LONGITUDE]),
        )
        self.distance = float(data[ATTR_DISTANCE])
        self.id = str(data[AEMET_ATTR_IDEMA])
        self.name = str(data[AEMET_ATTR_STATION_LOCATION])

        self.zoneinfo = timezone_from_coords(self.coords)

        self.update_sample(data)

    def get_altitude(self) -> float:
        """Return Station altitude."""
        return self.altitude

    def get_coords(self) -> tuple[float, float]:
        """Return Station coordinates."""
        return self.coords

    def get_datetime(self) -> datetime:
        """Return Station datetime of data."""
        return self._datetime

    def get_distance(self) -> float:
        """Return Station distance from selected coordinates."""
        return round(self.distance, 3)

    def get_dew_point(self) -> float:
        """Return Station dew point."""
        return self.dew_point

    def get_humidity(self) -> float:
        """Return Station humidity."""
        return self.humidity

    def get_id(self) -> str:
        """Return Station ID."""
        return self.id

    def get_name(self) -> str:
        """Return Station name."""
        return self.name

    def get_outdated(self) -> bool:
        """Return Station data outdated."""
        cur_dt = get_current_datetime()
        return cur_dt > self.get_datetime() + STATION_MAX_DELTA

    def get_precipitation(self) -> float:
        """Return Station precipitation."""
        return self.precipitation

    def get_pressure(self) -> float:
        """Return Station pressure."""
        return self.pressure

    def get_temp(self) -> float:
        """Return Station temperature."""
        return self.temp

    def get_temp_max(self) -> float:
        """Return Station maximum temperature."""
        return self.temp_max

    def get_temp_min(self) -> float:
        """Return Station minimum temperature."""
        return self.temp_min

    def get_timestamp(self) -> str:
        """Return Station timestamp."""
        return self._datetime.isoformat()

    def get_timezone(self) -> ZoneInfo:
        """Return Station timezone."""
        return self.zoneinfo

    def get_wind_direction(self) -> float | None:
        """Return Station wind direction."""
        return self.wind_direction

    def get_wind_speed(self) -> float | None:
        """Return Station wind speed."""
        return self.wind_speed

    def get_wind_speed_max(self) -> float | None:
        """Return Station maximum wind speed."""
        return self.wind_speed_max

    def update_sample(self, data: dict[str, Any]) -> None:
        """Update Station data from sample."""
        station_dt = parse_api_timestamp(data[AEMET_ATTR_STATION_DATE])

        self._datetime = station_dt.astimezone(self.get_timezone())
        self.dew_point = float(data[AEMET_ATTR_STATION_DEWPOINT])
        self.humidity = float(data[AEMET_ATTR_STATION_HUMIDITY])
        self.precipitation = float(data[AEMET_ATTR_STATION_PRECIPITATION])
        if AEMET_ATTR_STATION_PRESSURE_SEA in data:
            self.pressure = float(data[AEMET_ATTR_STATION_PRESSURE_SEA])
        else:
            self.pressure = float(data[AEMET_ATTR_STATION_PRESSURE])
        self.temp = float(data[AEMET_ATTR_STATION_TEMPERATURE])
        self.temp_max = float(data[AEMET_ATTR_STATION_TEMPERATURE_MAX])
        self.temp_min = float(data[AEMET_ATTR_STATION_TEMPERATURE_MIN])
        if AEMET_ATTR_STATION_WIND_DIRECTION in data:
            self.wind_direction = float(data[AEMET_ATTR_STATION_WIND_DIRECTION])
        if AEMET_ATTR_STATION_WIND_SPEED in data:
            self.wind_speed = float(data[AEMET_ATTR_STATION_WIND_SPEED])
        if AEMET_ATTR_STATION_WIND_SPEED_MAX in data:
            self.wind_speed_max = float(data[AEMET_ATTR_STATION_WIND_SPEED_MAX])

    def update_samples(self, samples: dict[str, Any]) -> None:
        """Update Station data from samples."""
        latest: dict[str, Any]
        latest_dt: datetime | None = None
        for sample in samples[ATTR_DATA]:
            dt = parse_api_timestamp(sample[AEMET_ATTR_STATION_DATE])
            if latest_dt is None or dt > latest_dt:
                latest = sample
                latest_dt = dt
        if (
            latest_dt is not None
            and self.get_datetime() < latest_dt <= get_current_datetime()
        ):
            self.update_sample(latest)

    def data(self) -> dict[str, Any]:
        """Return station data."""
        data: dict[str, Any] = {
            AOD_ALTITUDE: self.get_altitude(),
            AOD_COORDS: self.get_coords(),
            AOD_DATETIME: self.get_datetime(),
            AOD_DISTANCE: self.get_distance(),
            AOD_HUMIDITY: self.get_humidity(),
            AOD_ID: self.get_id(),
            AOD_NAME: self.get_name(),
            AOD_OUTDATED: self.get_outdated(),
            AOD_PRECIPITATION: self.get_precipitation(),
            AOD_PRESSURE: self.get_pressure(),
            AOD_TEMP: self.get_temp(),
            AOD_TEMP_MAX: self.get_temp_max(),
            AOD_TEMP_MIN: self.get_temp_min(),
            AOD_TIMESTAMP: self.get_timestamp(),
            AOD_TIMEZONE: self.get_timezone(),
        }

        wind_direction = self.get_wind_direction()
        if wind_direction is not None:
            data[AOD_WIND_DIRECTION] = wind_direction

        wind_speed = self.get_wind_speed()
        if wind_speed is not None:
            data[AOD_WIND_SPEED] = wind_speed

        wind_speed_max = self.get_wind_speed_max()
        if wind_speed_max is not None:
            data[AOD_WIND_SPEED_MAX] = wind_speed_max

        return data

    def weather(self) -> dict[str, Any]:
        """Return Station weather data."""
        weather: dict[str, Any] = {
            AOD_DEW_POINT: self.get_dew_point(),
            AOD_HUMIDITY: self.get_humidity(),
            AOD_PRECIPITATION: self.get_precipitation(),
            AOD_PRESSURE: self.get_pressure(),
            AOD_TEMP: self.get_temp(),
        }

        wind_direction = self.get_wind_direction()
        if wind_direction is not None:
            weather[AOD_WIND_DIRECTION] = wind_direction

        wind_speed = self.get_wind_speed()
        if wind_speed is not None:
            weather[AOD_WIND_SPEED] = wind_speed

        wind_speed_max = self.get_wind_speed_max()
        if wind_speed_max is not None:
            weather[AOD_WIND_SPEED_MAX] = wind_speed_max

        return weather



"""AEMET OpenData Constants."""

from datetime import timedelta
from typing import Final

AEMET_ATTR_DATA: Final[str] = "datos"
AEMET_ATTR_DATE: Final[str] = "fecha"
AEMET_ATTR_DAY: Final[str] = "dia"
AEMET_ATTR_DIRECTION: Final[str] = "direccion"
AEMET_ATTR_ELABORATED: Final[str] = "elaborado"
AEMET_ATTR_FEEL_TEMPERATURE: Final[str] = "sensTermica"
AEMET_ATTR_FORECAST: Final[str] = "prediccion"
AEMET_ATTR_GUST: Final[str] = "rachaMax"
AEMET_ATTR_HUMIDITY: Final[str] = "humedadRelativa"
AEMET_ATTR_ID: Final[str] = "id"
AEMET_ATTR_IDEMA: Final[str] = "idema"
AEMET_ATTR_MAX: Final[str] = "maxima"
AEMET_ATTR_MIN: Final[str] = "minima"
AEMET_ATTR_NAME: Final[str] = "nombre"
AEMET_ATTR_PERIOD: Final[str] = "periodo"
AEMET_ATTR_PRECIPITATION: Final[str] = "precipitacion"
AEMET_ATTR_PRECIPITATION_PROBABILITY: Final[str] = "probPrecipitacion"
AEMET_ATTR_SKY_STATE: Final[str] = "estadoCielo"
AEMET_ATTR_SNOW: Final[str] = "nieve"
AEMET_ATTR_SNOW_PROBABILITY: Final[str] = "probNieve"
AEMET_ATTR_SPEED: Final[str] = "velocidad"
AEMET_ATTR_STATION_ALTITUDE: Final[str] = "alt"
AEMET_ATTR_STATION_DATE: Final[str] = "fint"
AEMET_ATTR_STATION_DEWPOINT: Final[str] = "tpr"
AEMET_ATTR_STATION_HUMIDITY: Final[str] = "hr"
AEMET_ATTR_STATION_LATITUDE: Final[str] = "lat"
AEMET_ATTR_STATION_LOCATION: Final[str] = "ubi"
AEMET_ATTR_STATION_LONGITUDE: Final[str] = "lon"
AEMET_ATTR_STATION_PRECIPITATION: Final[str] = "prec"
AEMET_ATTR_STATION_PRESSURE: Final[str] = "pres"
AEMET_ATTR_STATION_PRESSURE_SEA: Final[str] = "pres_nmar"
AEMET_ATTR_STATION_TEMPERATURE: Final[str] = "ta"
AEMET_ATTR_STATION_TEMPERATURE_MAX: Final[str] = "tamax"
AEMET_ATTR_STATION_TEMPERATURE_MIN: Final[str] = "tamin"
AEMET_ATTR_STATION_WIND_DIRECTION: Final[str] = "dv"
AEMET_ATTR_STATION_WIND_SPEED: Final[str] = "vv"
AEMET_ATTR_STATION_WIND_SPEED_MAX: Final[str] = "vmax"
AEMET_ATTR_STATE: Final[str] = "estado"
AEMET_ATTR_STORM_PROBABILITY: Final[str] = "probTormenta"
AEMET_ATTR_SUN_RISE: Final[str] = "orto"
AEMET_ATTR_SUN_SET: Final[str] = "ocaso"
AEMET_ATTR_TEMPERATURE: Final[str] = "temperatura"
AEMET_ATTR_TOWN_ALTITUDE: Final[str] = "altitud"
AEMET_ATTR_TOWN_RESIDENTS: Final[str] = "num_hab"
AEMET_ATTR_TOWN_LATITUDE: Final[str] = "latitud"
AEMET_ATTR_TOWN_LATITUDE_DECIMAL: Final[str] = "latitud_dec"
AEMET_ATTR_TOWN_LONGITUDE: Final[str] = "longitud"
AEMET_ATTR_TOWN_LONGITUDE_DECIMAL: Final[str] = "longitud_dec"
AEMET_ATTR_UV_MAX: Final[str] = "uvMax"
AEMET_ATTR_VALUE: Final[str] = "value"
AEMET_ATTR_WEATHER_STATION_LATITUDE: Final[str] = "latitud"
AEMET_ATTR_WEATHER_STATION_LONGITUDE: Final[str] = "longitud"
AEMET_ATTR_WIND: Final[str] = "viento"
AEMET_ATTR_WIND_GUST: Final[str] = "vientoAndRachaMax"

AOD_ALTITUDE: Final[str] = "altitude"
AOD_COND_CLEAR_NIGHT: Final[str] = "clear-night"
AOD_COND_CLOUDY: Final[str] = "cloudy"
AOD_COND_FOG: Final[str] = "fog"
AOD_COND_LIGHTNING: Final[str] = "lightning"
AOD_COND_LIGHTNING_RAINY: Final[str] = "lightning-rainy"
AOD_COND_PARTLY_CLODUY: Final[str] = "partly-cloudy"
AOD_COND_POURING: Final[str] = "pouring"
AOD_COND_RAINY: Final[str] = "rainy"
AOD_COND_SNOWY: Final[str] = "snowy"
AOD_COND_SUNNY: Final[str] = "sunny"
AOD_CONDITION: Final[str] = "condition"
AOD_COORDS: Final[str] = "coordinates"
AOD_DATETIME: Final[str] = "datetime"
AOD_DEW_POINT: Final[str] = "dew-point"
AOD_DISTANCE: Final[str] = "distance"
AOD_FEEL_TEMP: Final[str] = "feel-temperature"
AOD_FEEL_TEMP_MAX: Final[str] = "feel-temperature-max"
AOD_FEEL_TEMP_MIN: Final[str] = "feel-temperature-min"
AOD_FORECAST: Final[str] = "forecast"
AOD_FORECAST_CURRENT: Final[str] = "forecast-current"
AOD_FORECAST_DAILY: Final[str] = "forecast-daily"
AOD_FORECAST_HOURLY: Final[str] = "forecast-hourly"
AOD_HUMIDITY: Final[str] = "humidity"
AOD_HUMIDITY_MAX: Final[str] = "humidity-max"
AOD_HUMIDITY_MIN: Final[str] = "humidity-min"
AOD_ID: Final[str] = "id"
AOD_NAME: Final[str] = "name"
AOD_OUTDATED: Final[str] = "outdated"
AOD_RAIN: Final[str] = "rain"
AOD_RAIN_PROBABILITY: Final[str] = "rain-probability"
AOD_PRECIPITATION: Final[str] = "precipitation"
AOD_PRECIPITATION_PROBABILITY: Final[str] = "precipitation-probability"
AOD_PRESSURE: Final[str] = "pressure"
AOD_RESIDENTS: Final[str] = "residents"
AOD_SNOW: Final[str] = "snow"
AOD_SNOW_PROBABILITY: Final[str] = "snow-probability"
AOD_STATION: Final[str] = "station"
AOD_STORM_PROBABILITY: Final[str] = "storm-probability"
AOD_SUNRISE: Final[str] = "sunrise"
AOD_SUNSET: Final[str] = "sunset"
AOD_TEMP: Final[str] = "temperature"
AOD_TEMP_FEELING: Final[str] = "temperature-feeling"
AOD_TEMP_MAX: Final[str] = "temperature-max"
AOD_TEMP_MIN: Final[str] = "temperature-min"
AOD_TIMESTAMP: Final[str] = "timestamp"
AOD_TIMEZONE: Final[str] = "timezone"
AOD_TOWN: Final[str] = "town"
AOD_UV_INDEX: Final[str] = "uv-index"
AOD_WEATHER: Final[str] = "weather"
AOD_WIND_DIRECTION: Final[str] = "wind-direction"
AOD_WIND_SPEED: Final[str] = "wind-speed"
AOD_WIND_SPEED_MAX: Final[str] = "wind-speed-max"

API_ID_PFX: Final[str] = "id"
API_MIN_STATION_DISTANCE_KM: Final[int] = 40
API_MIN_TOWN_DISTANCE_KM: Final[int] = 40
API_PERIOD_24H: Final[int] = 24
API_PERIOD_FULL_DAY: Final[str] = "00-24"
API_PERIOD_HALF_1_DAY: Final[str] = "00-12"
API_PERIOD_HALF_2_DAY: Final[str] = "12-24"
API_PERIOD_QUARTER_1_DAY: Final[str] = "00-06"
API_PERIOD_QUARTER_2_DAY: Final[str] = "06-12"
API_PERIOD_QUARTER_3_DAY: Final[str] = "12-18"
API_PERIOD_QUARTER_4_DAY: Final[str] = "18-24"
API_PERIOD_SPLIT: Final[int] = 2
API_TIMEOUT: Final[int] = 15
API_URL: Final[str] = "https://opendata.aemet.es/opendata/api"

ATTR_DATA: Final[str] = "data"
ATTR_DISTANCE: Final[str] = "distance"
ATTR_RESPONSE: Final[str] = "response"

RAW_FORECAST_DAILY: Final[str] = "forecast-daily"
RAW_FORECAST_HOURLY: Final[str] = "forecast-hourly"
RAW_STATIONS: Final[str] = "stations"
RAW_TOWNS: Final[str] = "towns"

STATION_MAX_DELTA: Final[timedelta] = timedelta(hours=2)



"""Client for the AEMET OpenData REST API."""

import asyncio
from dataclasses import dataclass, field
import logging
from typing import Any, cast

from aiohttp import ClientError, ClientSession
from aiohttp.client_reqrep import ClientResponse
import geopy.distance
from geopy.distance import Distance

from .const import (
    AEMET_ATTR_DATA,
    AEMET_ATTR_STATE,
    AEMET_ATTR_STATION_LATITUDE,
    AEMET_ATTR_STATION_LONGITUDE,
    AEMET_ATTR_TOWN_LATITUDE_DECIMAL,
    AEMET_ATTR_TOWN_LONGITUDE_DECIMAL,
    AEMET_ATTR_WEATHER_STATION_LATITUDE,
    AEMET_ATTR_WEATHER_STATION_LONGITUDE,
    AOD_CONDITION,
    AOD_DEW_POINT,
    AOD_FEEL_TEMP,
    AOD_HUMIDITY,
    AOD_PRECIPITATION,
    AOD_PRECIPITATION_PROBABILITY,
    AOD_PRESSURE,
    AOD_RAIN,
    AOD_RAIN_PROBABILITY,
    AOD_SNOW,
    AOD_SNOW_PROBABILITY,
    AOD_STATION,
    AOD_STORM_PROBABILITY,
    AOD_TEMP,
    AOD_TIMESTAMP,
    AOD_TOWN,
    AOD_UV_INDEX,
    AOD_WEATHER,
    AOD_WIND_DIRECTION,
    AOD_WIND_SPEED,
    AOD_WIND_SPEED_MAX,
    API_MIN_STATION_DISTANCE_KM,
    API_MIN_TOWN_DISTANCE_KM,
    API_TIMEOUT,
    API_URL,
    ATTR_DATA,
    ATTR_DISTANCE,
    ATTR_RESPONSE,
    RAW_FORECAST_DAILY,
    RAW_FORECAST_HOURLY,
    RAW_STATIONS,
    RAW_TOWNS,
)
from .exceptions import (
    AemetError,
    ApiError,
    AuthError,
    StationNotFound,
    TooManyRequests,
    TownNotFound,
)
from .helpers import get_current_datetime, parse_station_coordinates, parse_town_code
from .station import Station
from .town import Town

_LOGGER = logging.getLogger(__name__)


@dataclass
class ConnectionOptions:
    """AEMET OpenData API options for connection."""

    api_key: str
    station_data: bool = False


@dataclass
class LegacyWeather:
    """Legacy class for AEMET weather data."""

    daily: dict[str, Any] = field(default_factory=dict)
    hourly: dict[str, Any] = field(default_factory=dict)
    station: dict[str, Any] = field(default_factory=dict)


class AEMET:
    """Interacts with the AEMET OpenData API."""

    _api_raw_data: dict[str, Any]
    aiohttp_session: ClientSession
    coords: tuple[float, float] | None
    dist_hp: bool
    headers: dict[str, Any]
    legacy: dict[str, Any]
    options: ConnectionOptions
    station: Station | None
    town: Town | None

    def __init__(
        self,
        aiohttp_session: ClientSession,
        options: ConnectionOptions,
    ) -> None:
        """Init AEMET OpenData API."""
        self._api_raw_data = {
            RAW_FORECAST_DAILY: {},
            RAW_FORECAST_HOURLY: {},
            RAW_STATIONS: {},
            RAW_TOWNS: {},
        }
        self.aiohttp_session = aiohttp_session
        self.coords = None
        self.dist_hp = False
        self.headers = {
            "Cache-Control": "no-cache",
            "api_key": options.api_key,
        }
        self.legacy = {
            "daily": None,
            "hourly": None,
            "station": None,
        }
        self.options = options
        self.station = None
        self.town = None

    async def api_call(self, cmd: str, fetch_data: bool = False) -> dict[str, Any]:
        """Perform Rest API call."""
        _LOGGER.debug("api_call: cmd=%s", cmd)

        try:
            resp: ClientResponse = await self.aiohttp_session.request(
                "GET",
                f"{API_URL}/{cmd}",
                timeout=API_TIMEOUT,
                headers=self.headers,
            )
        except ClientError as err:
            raise AemetError(err) from err

        if resp.status == 401:
            raise AuthError("API authentication error")
        if resp.status == 404:
            raise ApiError("API data error")
        if resp.status == 429:
            raise TooManyRequests("Too many API requests")
        if resp.status != 200:
            raise AemetError(f"API status={resp.status}")

        resp_json = await resp.json(content_type=None)
        _LOGGER.debug("api_call: cmd=%s resp=%s", cmd, resp_json)

        if isinstance(resp_json, dict):
            if resp_json.get(AEMET_ATTR_STATE) == 404:
                raise ApiError("API data error")

        json_response = cast(dict[str, Any], resp_json)
        if fetch_data and AEMET_ATTR_DATA in json_response:
            data = await self.api_data(json_response[AEMET_ATTR_DATA])
            if data:
                json_response = {
                    ATTR_RESPONSE: json_response,
                    ATTR_DATA: data,
                }
        if isinstance(json_response, list):
            json_response = {
                ATTR_DATA: json_response,
            }

        return json_response

    async def api_data(self, url: str) -> dict[str, Any]:
        """Fetch API data."""
        _LOGGER.debug("api_data: url=%s", url)

        try:
            resp: ClientResponse = await self.aiohttp_session.request(
                "GET",
                url,
                timeout=API_TIMEOUT,
            )
        except ClientError as err:
            raise AemetError(err) from err

        if resp.status == 404:
            raise ApiError("API data error")
        if resp.status == 429:
            raise TooManyRequests("Too many API requests")
        if resp.status != 200:
            raise AemetError(f"API status={resp.status}")

        resp_json = await resp.json(content_type=None)
        _LOGGER.debug("api_data: url=%s resp=%s", url, resp_json)

        if isinstance(resp_json, dict):
            if resp_json.get(AEMET_ATTR_STATE) == 404:
                raise ApiError("API data error")

        return cast(dict[str, Any], resp_json)

    def legacy_weather(self) -> LegacyWeather:
        """Return legacy weather data."""
        return LegacyWeather(
            self.legacy["daily"],
            self.legacy["hourly"],
            self.legacy["station"],
        )

    def raw_data(self) -> dict[str, Any]:
        """Return raw AEMET OpenData API data."""
        return self._api_raw_data

    def data(self) -> dict[str, Any]:
        """Return AEMET OpenData data."""
        data: dict[str, Any] = {}

        if self.station is not None:
            data[AOD_STATION] = self.station.data()

        if self.town is not None:
            data[AOD_TOWN] = self.town.data()

        weather = self.weather()
        if weather is not None:
            data[AOD_WEATHER] = weather

        data[AOD_TIMESTAMP] = get_current_datetime().isoformat()

        return data

    def calc_distance(
        self, start: tuple[float, float], end: tuple[float, float]
    ) -> Distance:
        """Calculate distance between 2 points."""
        if self.dist_hp:
            return geopy.distance.geodesic(start, end)
        return geopy.distance.great_circle(start, end)

    def distance_high_precision(self, dist_hp: bool) -> bool:
        """Enable/Disable high precision for distance calculations."""
        self.dist_hp = dist_hp
        return self.dist_hp

    async def get_climatological_values_stations(
        self, fetch_data: bool = True
    ) -> dict[str, Any]:
        """Get stations available for climatological values."""
        return await self.api_call(
            "valores/climatologicos/inventarioestaciones/todasestaciones", fetch_data
        )

    async def get_climatological_values_station_by_coordinates(
        self, latitude: float, longitude: float
    ) -> dict[str, Any]:
        """Get closest climatological values station to coordinates."""
        station: dict[str, Any] | None = None
        stations = await self.get_climatological_values_stations()
        search_coords = (latitude, longitude)
        distance = API_MIN_STATION_DISTANCE_KM
        for cur_station in stations[ATTR_DATA]:
            station_coords = parse_station_coordinates(
                cur_station[AEMET_ATTR_WEATHER_STATION_LATITUDE],
                cur_station[AEMET_ATTR_WEATHER_STATION_LONGITUDE],
            )
            station_point = geopy.point.Point(station_coords)
            cur_coords = (station_point.latitude, station_point.longitude)
            cur_distance = self.calc_distance(search_coords, cur_coords).km
            if cur_distance < distance:
                distance = cur_distance
                station = cur_station
        if station is None:
            raise StationNotFound(f"No stations found for [{latitude}, {longitude}]")
        _LOGGER.debug("distance: %s, station: %s", distance, station)
        return station

    async def get_climatological_values_station_data(
        self, station: str, fetch_data: bool = True
    ) -> dict[str, Any]:
        """Get data from climatological values station."""
        return await self.api_call(
            f"valores/climatologicos/inventarioestaciones/estaciones/{station}",
            fetch_data,
        )

    async def get_conventional_observation_stations(
        self, fetch_data: bool = True
    ) -> dict[str, Any]:
        """Get stations available for conventional observations."""
        return await self.api_call("observacion/convencional/todas", fetch_data)

    async def get_conventional_observation_station_by_coordinates(
        self, latitude: float, longitude: float
    ) -> dict[str, Any]:
        """Get closest conventional observation station to coordinates."""
        station: dict[str, Any] | None = None
        stations = await self.get_conventional_observation_stations()
        search_coords = (latitude, longitude)
        distance = API_MIN_STATION_DISTANCE_KM
        for cur_station in stations[ATTR_DATA]:
            cur_coords = (
                cur_station[AEMET_ATTR_STATION_LATITUDE],
                cur_station[AEMET_ATTR_STATION_LONGITUDE],
            )
            cur_distance = self.calc_distance(search_coords, cur_coords).km
            if cur_distance < distance:
                distance = cur_distance
                station = cur_station
        if station is None:
            raise StationNotFound(f"No stations found for [{latitude}, {longitude}]")
        _LOGGER.debug("distance: %s, station: %s", distance, station)
        station[ATTR_DISTANCE] = distance
        return station

    async def get_conventional_observation_station_data(
        self, station: str, fetch_data: bool = True
    ) -> dict[str, Any]:
        """Get data from conventional observation station."""
        res = await self.api_call(
            f"observacion/convencional/datos/estacion/{station}", fetch_data
        )
        self._api_raw_data[RAW_STATIONS][station] = res
        return res

    async def get_lightnings_map(self) -> dict[str, Any]:
        """Get map with lightning falls (last 6h)."""
        return await self.api_call("red/rayos/mapa")

    async def get_specific_forecast_town_daily(
        self, town: str, fetch_data: bool = True
    ) -> dict[str, Any]:
        """Get daily forecast for specific town (daily)."""
        res = await self.api_call(
            f"prediccion/especifica/municipio/diaria/{parse_town_code(town)}",
            fetch_data,
        )
        self._api_raw_data[RAW_FORECAST_DAILY][town] = res
        return res

    async def get_specific_forecast_town_hourly(
        self, town: str, fetch_data: bool = True
    ) -> dict[str, Any]:
        """Get hourly forecast for specific town (hourly)."""
        res = await self.api_call(
            f"prediccion/especifica/municipio/horaria/{parse_town_code(town)}",
            fetch_data,
        )
        self._api_raw_data[RAW_FORECAST_HOURLY][town] = res
        return res

    async def get_town(self, town: str) -> dict[str, Any]:
        """Get information about specific town."""
        res = await self.api_call(f"maestro/municipio/{town}")
        self._api_raw_data[RAW_TOWNS][town] = res
        return res

    async def get_town_by_coordinates(
        self, latitude: float, longitude: float
    ) -> dict[str, Any]:
        """Get closest town to coordinates."""
        town: dict[str, Any] | None = None
        towns = await self.get_towns()
        search_coords = (latitude, longitude)
        distance = API_MIN_TOWN_DISTANCE_KM
        for cur_town in towns[ATTR_DATA]:
            cur_coords = (
                cur_town[AEMET_ATTR_TOWN_LATITUDE_DECIMAL],
                cur_town[AEMET_ATTR_TOWN_LONGITUDE_DECIMAL],
            )
            cur_distance = self.calc_distance(search_coords, cur_coords).km
            if cur_distance < distance:
                distance = cur_distance
                town = cur_town
        if town is None:
            raise TownNotFound(f"No towns found for [{latitude}, {longitude}]")
        _LOGGER.debug("distance: %s, town: %s", distance, town)
        town[ATTR_DISTANCE] = distance
        return town

    async def get_towns(self) -> dict[str, Any]:
        """Get information about towns."""
        return await self.api_call("maestro/municipios")

    async def select_coordinates(self, latitude: float, longitude: float) -> None:
        """Select town and station based on provided coordinates."""
        coords = (latitude, longitude)

        if self.options.station_data:
            try:
                station_data = (
                    await self.get_conventional_observation_station_by_coordinates(
                        latitude,
                        longitude,
                    )
                )
            except StationNotFound as err:
                _LOGGER.error(err)
                station_data = None
        else:
            station_data = None

        town_data = await self.get_town_by_coordinates(latitude, longitude)

        self.coords = coords
        if station_data is not None:
            self.station = Station(station_data)
        self.town = Town(town_data)

    async def update_daily(self) -> None:
        """Update AEMET OpenData town daily forecast."""
        if self.town is not None:
            town_id = self.town.get_id()
            daily = await self.get_specific_forecast_town_daily(town_id)
            self.legacy["daily"] = daily
            self.town.update_daily(daily)

    async def update_hourly(self) -> None:
        """Update AEMET OpenData town hourly forecast."""
        if self.town is not None:
            town_id = self.town.get_id()
            hourly = await self.get_specific_forecast_town_hourly(town_id)
            self.legacy["hourly"] = hourly
            self.town.update_hourly(hourly)

    async def update_station(self) -> None:
        """Update AEMET OpenData station."""
        if self.station is not None:
            station_id = self.station.get_id()
            station = await self.get_conventional_observation_station_data(station_id)
            self.legacy["station"] = station
            self.station.update_samples(station)

    async def update(self) -> None:
        """Update all AEMET OpenData data."""
        tasks = [
            self.update_daily(),
            self.update_hourly(),
            self.update_station(),
        ]
        await asyncio.gather(*tasks)

    def weather(self) -> dict[str, Any] | None:
        """Update AEMET OpenData town daily forecast."""
        daily: dict[str, Any]
        hourly: dict[str, Any]
        station: dict[str, Any]

        if self.station is not None and not self.station.get_outdated():
            station = self.station.weather()
        else:
            station = {}

        if self.town is not None:
            daily = self.town.weather_daily()
            hourly = self.town.weather_hourly()
        else:
            daily = {}
            hourly = {}

        condition = hourly.get(AOD_CONDITION) or daily.get(AOD_CONDITION)
        dew_point = station.get(AOD_DEW_POINT)
        feel_temp = hourly.get(AOD_FEEL_TEMP)
        humidity = station.get(AOD_HUMIDITY)
        if humidity is None:
            humidity = hourly.get(AOD_HUMIDITY)
        pressure = station.get(AOD_PRESSURE)
        precipitation = station.get(AOD_PRECIPITATION)
        if precipitation is None:
            precipitation = hourly.get(AOD_PRECIPITATION)
        precipitation_prob = hourly.get(AOD_PRECIPITATION_PROBABILITY)
        if precipitation_prob is None:
            precipitation_prob = daily.get(AOD_PRECIPITATION_PROBABILITY)
        rain = station.get(AOD_PRECIPITATION)
        if rain is None:
            rain = hourly.get(AOD_RAIN)
        rain_prob = hourly.get(AOD_RAIN_PROBABILITY)
        snow = hourly.get(AOD_SNOW)
        snow_prob = hourly.get(AOD_SNOW_PROBABILITY)
        storm_prob = hourly.get(AOD_STORM_PROBABILITY)
        temp = station.get(AOD_TEMP)
        if temp is None:
            temp = hourly.get(AOD_TEMP)
        uv_index = daily.get(AOD_UV_INDEX)
        wind_direction = (
            station.get(AOD_WIND_DIRECTION)
            or hourly.get(AOD_WIND_DIRECTION)
            or daily.get(AOD_WIND_DIRECTION)
        )
        wind_speed = (
            station.get(AOD_WIND_SPEED)
            or hourly.get(AOD_WIND_SPEED)
            or daily.get(AOD_WIND_SPEED)
        )
        wind_speed_max = station.get(AOD_WIND_SPEED_MAX) or hourly.get(
            AOD_WIND_SPEED_MAX
        )

        weather: dict[str, Any] = {
            AOD_CONDITION: condition,
            AOD_DEW_POINT: dew_point,
            AOD_HUMIDITY: humidity,
            AOD_FEEL_TEMP: feel_temp,
            AOD_PRECIPITATION: precipitation,
            AOD_PRECIPITATION_PROBABILITY: precipitation_prob,
            AOD_PRESSURE: pressure,
            AOD_RAIN: rain,
            AOD_RAIN_PROBABILITY: rain_prob,
            AOD_SNOW: snow,
            AOD_SNOW_PROBABILITY: snow_prob,
            AOD_STORM_PROBABILITY: storm_prob,
            AOD_TEMP: temp,
            AOD_UV_INDEX: uv_index,
            AOD_WIND_DIRECTION: wind_direction,
            AOD_WIND_SPEED: wind_speed,
            AOD_WIND_SPEED_MAX: wind_speed_max,
        }

        return weather



"""AEMET OpenData Forecast."""

from datetime import datetime
from typing import Any, Final

from .const import (
    AEMET_ATTR_DIRECTION,
    AEMET_ATTR_FEEL_TEMPERATURE,
    AEMET_ATTR_HUMIDITY,
    AEMET_ATTR_MAX,
    AEMET_ATTR_MIN,
    AEMET_ATTR_PERIOD,
    AEMET_ATTR_PRECIPITATION,
    AEMET_ATTR_PRECIPITATION_PROBABILITY,
    AEMET_ATTR_SKY_STATE,
    AEMET_ATTR_SNOW,
    AEMET_ATTR_SNOW_PROBABILITY,
    AEMET_ATTR_SPEED,
    AEMET_ATTR_STORM_PROBABILITY,
    AEMET_ATTR_SUN_RISE,
    AEMET_ATTR_SUN_SET,
    AEMET_ATTR_TEMPERATURE,
    AEMET_ATTR_UV_MAX,
    AEMET_ATTR_VALUE,
    AEMET_ATTR_WIND,
    AEMET_ATTR_WIND_GUST,
    AOD_COND_CLEAR_NIGHT,
    AOD_COND_CLOUDY,
    AOD_COND_FOG,
    AOD_COND_LIGHTNING,
    AOD_COND_LIGHTNING_RAINY,
    AOD_COND_PARTLY_CLODUY,
    AOD_COND_POURING,
    AOD_COND_RAINY,
    AOD_COND_SNOWY,
    AOD_COND_SUNNY,
    AOD_CONDITION,
    AOD_FEEL_TEMP,
    AOD_FEEL_TEMP_MAX,
    AOD_FEEL_TEMP_MIN,
    AOD_HUMIDITY,
    AOD_HUMIDITY_MAX,
    AOD_HUMIDITY_MIN,
    AOD_PRECIPITATION,
    AOD_PRECIPITATION_PROBABILITY,
    AOD_RAIN,
    AOD_RAIN_PROBABILITY,
    AOD_SNOW,
    AOD_SNOW_PROBABILITY,
    AOD_STORM_PROBABILITY,
    AOD_SUNRISE,
    AOD_SUNSET,
    AOD_TEMP,
    AOD_TEMP_MAX,
    AOD_TEMP_MIN,
    AOD_TIMESTAMP,
    AOD_UV_INDEX,
    AOD_WIND_DIRECTION,
    AOD_WIND_SPEED,
    AOD_WIND_SPEED_MAX,
    API_PERIOD_24H,
    API_PERIOD_FULL_DAY,
    API_PERIOD_HALF_2_DAY,
    API_PERIOD_SPLIT,
)

API_CONDITIONS_MAP: Final[dict[str, list[str]]] = {
    AOD_COND_CLEAR_NIGHT: [
        "11n",  # Despejado (de noche)
    ],
    AOD_COND_CLOUDY: [
        "14",  # Nuboso
        "14n",  # Nuboso (de noche)
        "15",  # Muy nuboso
        "15n",  # Muy nuboso (de noche)
        "16",  # Cubierto
        "16n",  # Cubierto (de noche)
        "17",  # Nubes altas
        "17n",  # Nubes altas (de noche)
    ],
    AOD_COND_FOG: [
        "81",  # Niebla
        "81n",  # Niebla (de noche)
        "82",  # Bruma - Neblina
        "82n",  # Bruma - Neblina (de noche)
    ],
    AOD_COND_LIGHTNING: [
        "51",  # Intervalos nubosos con tormenta
        "51n",  # Intervalos nubosos con tormenta (de noche)
        "52",  # Nuboso con tormenta
        "52n",  # Nuboso con tormenta (de noche)
        "53",  # Muy nuboso con tormenta
        "53n",  # Muy nuboso con tormenta (de noche)
        "54",  # Cubierto con tormenta
        "54n",  # Cubierto con tormenta (de noche)
    ],
    AOD_COND_LIGHTNING_RAINY: [
        "61",  # Intervalos nubosos con tormenta y lluvia escasa
        "61n",  # Intervalos nubosos con tormenta y lluvia escasa (de noche)
        "62",  # Nuboso con tormenta y lluvia escasa
        "62n",  # Nuboso con tormenta y lluvia escasa (de noche)
        "63",  # Muy nuboso con tormenta y lluvia escasa
        "63n",  # Muy nuboso con tormenta y lluvia escasa (de noche)
        "64",  # Cubierto con tormenta y lluvia escasa
        "64n",  # Cubierto con tormenta y lluvia escasa (de noche)
    ],
    AOD_COND_PARTLY_CLODUY: [
        "12",  # Poco nuboso
        "12n",  # Poco nuboso (de noche)
        "13",  # Intervalos nubosos
        "13n",  # Intervalos nubosos (de noche)
    ],
    AOD_COND_POURING: [
        "27",  # Chubascos
        "27n",  # Chubascos (de noche)
    ],
    AOD_COND_RAINY: [
        "23",  # Intervalos nubosos con lluvia
        "23n",  # Intervalos nubosos con lluvia (de noche)
        "24",  # Nuboso con lluvia
        "24n",  # Nuboso con lluvia (de noche)
        "25",  # Muy nuboso con lluvia
        "25n",  # Muy nuboso con lluvia (de noche)
        "26",  # Cubierto con lluvia
        "26n",  # Cubierto con lluvia (de noche)
        "43",  # Intervalos nubosos con lluvia escasa
        "43n",  # Intervalos nubosos con lluvia escasa (de noche)
        "44",  # Nuboso con lluvia escasa
        "44n",  # Nuboso con lluvia escasa (de noche)
        "45",  # Muy nuboso con lluvia escasa
        "45n",  # Muy nuboso con lluvia escasa (de noche)
        "46",  # Cubierto con lluvia escasa
        "46n",  # Cubierto con lluvia escasa (de noche)
    ],
    AOD_COND_SNOWY: [
        "33",  # Intervalos nubosos con nieve
        "33n",  # Intervalos nubosos con nieve (de noche)
        "34",  # Nuboso con nieve
        "34n",  # Nuboso con nieve (de noche)
        "35",  # Muy nuboso con nieve
        "35n",  # Muy nuboso con nieve (de noche)
        "36",  # Cubierto con nieve
        "36n",  # Cubierto con nieve (de noche)
        "71",  # Intervalos nubosos con nieve escasa
        "71n",  # Intervalos nubosos con nieve escasa (de noche)
        "72",  # Nuboso con nieve escasa
        "72n",  # Nuboso con nieve escasa (de noche)
        "73",  # Muy nuboso con nieve escasa
        "73n",  # Muy nuboso con nieve escasa (de noche)
        "74",  # Cubierto con nieve escasa
        "74n",  # Cubierto con nieve escasa (de noche)
    ],
    AOD_COND_SUNNY: [
        "11",  # Despejado
    ],
}


WIND_DIRECTION_MAP: Final[dict[str, float | None]] = {
    "C": None,
    "N": 0.0,
    "NE": 45.0,
    "E": 90.0,
    "SE": 135.0,
    "S": 180.0,
    "SO": 225.0,
    "O": 270.0,
    "NO": 315.0,
}


def hash_api_conditions(conditions_map: dict[str, list[str]]) -> dict[str, str]:
    """Hash API conditions for faster access."""
    res: dict[str, Any] = {}
    for k, v in conditions_map.items():
        for c in v:
            res[c] = k
    return res


CONDITIONS_DICT: Final[dict[str, str]] = hash_api_conditions(API_CONDITIONS_MAP)


class ForecastValue:
    """AEMET OpenData Town Forecast value."""

    @classmethod
    def parse_condition(cls, condition: str) -> str:
        """Parse forecast condition from API to human readable."""
        return CONDITIONS_DICT.get(condition, condition)

    @classmethod
    def parse_precipitation(cls, precipitation: str | None) -> float | None:
        """Parse forecast precipitation into a float."""
        if precipitation is None:
            return None
        if precipitation == "Ip":
            return 0.0
        return float(precipitation)

    @classmethod
    def parse_wind_direction(cls, wind_direction: str) -> float | None:
        """Parse forecast wind direction into a float."""
        return WIND_DIRECTION_MAP.get(wind_direction)


class DailyForecastValue(ForecastValue):
    """AEMET OpenData Town Daily Forecast value."""

    condition: str
    _datetime: datetime
    feel_temp_max: int
    feel_temp_min: int
    humidity_max: int
    humidity_min: int
    periods: list[str] = [
        API_PERIOD_FULL_DAY,
        API_PERIOD_HALF_2_DAY,
    ]
    precipitation_prob: int
    temp_max: int
    temp_min: int
    uv_index: int | None
    wind_direction: float | None
    wind_speed: float | None

    def __init__(self, data: dict[str, Any], dt: datetime) -> None:
        """Init AEMET OpenData Town Daily Forecast."""
        condition = self.parse_value(data[AEMET_ATTR_SKY_STATE])
        if condition is None or not condition:
            raise ValueError

        self.condition = self.parse_condition(condition)
        self._datetime = dt
        self.feel_temp_max = int(
            self.parse_value(data[AEMET_ATTR_FEEL_TEMPERATURE], key=AEMET_ATTR_MAX)
        )
        self.feel_temp_min = int(
            self.parse_value(data[AEMET_ATTR_FEEL_TEMPERATURE], key=AEMET_ATTR_MIN)
        )
        self.humidity_max = int(
            self.parse_value(data[AEMET_ATTR_HUMIDITY], key=AEMET_ATTR_MAX)
        )
        self.humidity_min = int(
            self.parse_value(data[AEMET_ATTR_HUMIDITY], key=AEMET_ATTR_MIN)
        )
        self.precipitation_prob = int(
            self.parse_value(data[AEMET_ATTR_PRECIPITATION_PROBABILITY])
        )
        self.temp_max = int(
            self.parse_value(data[AEMET_ATTR_TEMPERATURE], key=AEMET_ATTR_MAX)
        )
        self.temp_min = int(
            self.parse_value(data[AEMET_ATTR_TEMPERATURE], key=AEMET_ATTR_MIN)
        )
        self.wind_direction = self.parse_wind_direction(
            self.parse_value(data[AEMET_ATTR_WIND], key=AEMET_ATTR_DIRECTION),
        )

        if self.wind_direction is not None:
            self.wind_speed = int(
                self.parse_value(data[AEMET_ATTR_WIND], key=AEMET_ATTR_SPEED)
            )
        else:
            self.wind_speed = None

        if AEMET_ATTR_UV_MAX in data:
            self.uv_index = int(data[AEMET_ATTR_UV_MAX])
        else:
            self.uv_index = None

    def get_condition(self) -> str | None:
        """Return Town daily forecast condition."""
        return self.condition

    def get_datetime(self) -> datetime:
        """Return Town daily forecast datetime."""
        return self._datetime

    def get_feel_temp_max(self) -> int:
        """Return Town daily forecast maximum feel temperature."""
        return self.feel_temp_max

    def get_feel_temp_min(self) -> int:
        """Return Town daily forecast minimum feel temperature."""
        return self.feel_temp_min

    def get_humidity_max(self) -> int:
        """Return Town daily forecast maximum humidity."""
        return self.humidity_max

    def get_humidity_min(self) -> int:
        """Return Town daily forecast minimum humidity."""
        return self.humidity_min

    def get_precipitation_prob(self) -> int:
        """Return Town daily forecast precipitation probability."""
        return self.precipitation_prob

    def get_temp_max(self) -> int:
        """Return Town daily forecast maximum temperature."""
        return self.temp_max

    def get_temp_min(self) -> int:
        """Return Town daily forecast minimum temperature."""
        return self.temp_min

    def get_timestamp(self) -> str:
        """Return Town daily forecast timestamp."""
        return self._datetime.isoformat()

    def get_uv_index(self) -> int | None:
        """Return Town daily forecast UV index."""
        return self.uv_index

    def get_wind_direction(self) -> float | None:
        """Return Town daily forecast wind direction."""
        return self.wind_direction

    def get_wind_speed(self) -> float | None:
        """Return Town daily forecast wind speed."""
        return self.wind_speed

    def data(self) -> dict[str, Any]:
        """Return Town daily forecast data."""
        data: dict[str, Any] = {
            AOD_CONDITION: self.get_condition(),
            AOD_FEEL_TEMP_MAX: self.get_feel_temp_max(),
            AOD_FEEL_TEMP_MIN: self.get_feel_temp_min(),
            AOD_HUMIDITY_MAX: self.get_humidity_max(),
            AOD_HUMIDITY_MIN: self.get_humidity_min(),
            AOD_PRECIPITATION_PROBABILITY: self.get_precipitation_prob(),
            AOD_TEMP_MAX: self.get_temp_max(),
            AOD_TEMP_MIN: self.get_temp_min(),
            AOD_TIMESTAMP: self.get_timestamp(),
            AOD_UV_INDEX: self.get_uv_index(),
            AOD_WIND_DIRECTION: self.get_wind_direction(),
            AOD_WIND_SPEED: self.get_wind_speed(),
        }

        return data

    def parse_value(
        self, values: dict[str, Any] | list[Any], key: str = AEMET_ATTR_VALUE
    ) -> Any:
        """Parse Town daily forecast value from data."""
        if isinstance(values, list):
            if len(values) > 1:
                for value in values:
                    if key not in value:
                        continue
                    if isinstance(value[key], str) and not value[key]:
                        continue
                    for period in self.periods:
                        if value[AEMET_ATTR_PERIOD] == period:
                            return value[key]
            else:
                if key in values[0]:
                    return values[0][key]
        if isinstance(values, dict):
            if key in values:
                return values[key]
        return None


class HourlyForecastValue(ForecastValue):
    """AEMET OpenData Town Hourly Forecast value."""

    condition: str
    _datetime: datetime
    feel_temp: int | None
    humidity: int | None
    rain: float | None
    rain_probability: int | None
    snow: float | None
    snow_probability: int | None
    storm_probability: int | None
    sunrise: str
    sunset: str
    temp: int | None
    wind_direction: float | None
    wind_speed: float | None
    wind_speed_max: float | None

    def __init__(self, data: dict[str, Any], dt: datetime, hour: int) -> None:
        """Init AEMET OpenData Town Hourly Forecast."""
        condition = self.parse_value(data[AEMET_ATTR_SKY_STATE], hour)
        if condition is None:
            raise ValueError

        self.condition = self.parse_condition(condition)
        self._datetime = dt.replace(hour=hour)
        self.sunrise = str(data[AEMET_ATTR_SUN_RISE])
        self.sunset = str(data[AEMET_ATTR_SUN_SET])

        feel_temp = self.parse_value(data[AEMET_ATTR_FEEL_TEMPERATURE], hour)
        if feel_temp is not None:
            self.feel_temp = int(feel_temp)
        else:
            self.feel_temp = None

        humidity = self.parse_value(data[AEMET_ATTR_HUMIDITY], hour)
        if humidity is not None:
            self.humidity = int(humidity)
        else:
            self.humidity = None

        temp = self.parse_value(data[AEMET_ATTR_TEMPERATURE], hour)
        if temp is not None:
            self.temp = int(temp)
        else:
            self.temp = None

        rain_probability = self.parse_interval_value(
            data[AEMET_ATTR_PRECIPITATION_PROBABILITY], hour
        )
        if rain_probability is not None:
            self.rain_probability = int(rain_probability)

            rain = self.parse_value(data[AEMET_ATTR_PRECIPITATION], hour)
            if rain is not None:
                self.rain = self.parse_precipitation(rain)
            else:
                self.rain = None
        else:
            self.rain = None
            self.rain_probability = None

        snow_probability = self.parse_interval_value(
            data[AEMET_ATTR_SNOW_PROBABILITY], hour
        )
        if snow_probability is not None:
            self.snow_probability = int(snow_probability)

            snow = self.parse_value(data[AEMET_ATTR_SNOW], hour)
            if snow is not None:
                self.snow = self.parse_precipitation(snow)
            else:
                self.snow = None
        else:
            self.snow = None
            self.snow_probability = None

        storm_probability = self.parse_interval_value(
            data[AEMET_ATTR_STORM_PROBABILITY], hour
        )
        if storm_probability is not None:
            self.storm_probability = int(storm_probability)
        else:
            self.storm_probability = None

        wind_direction = self.parse_value(
            data[AEMET_ATTR_WIND_GUST],
            hour,
            key=AEMET_ATTR_DIRECTION,
        )
        if wind_direction is not None:
            self.wind_direction = self.parse_wind_direction(wind_direction[0])
        else:
            self.wind_direction = None

        if self.wind_direction is not None:
            self.wind_speed = float(
                self.parse_value(
                    data[AEMET_ATTR_WIND_GUST], hour, key=AEMET_ATTR_SPEED
                )[0]
            )
            self.wind_speed_max = float(
                self.parse_value(data[AEMET_ATTR_WIND_GUST], hour)
            )
        else:
            self.wind_speed = None
            self.wind_speed_max = None

    def get_condition(self) -> str:
        """Return Town hourly forecast condition."""
        return self.condition

    def get_datetime(self) -> datetime:
        """Return Town hourly forecast datetime."""
        return self._datetime

    def get_feel_temp(self) -> int | None:
        """Return Town hourly forecast feel temperature."""
        return self.feel_temp

    def get_humidity(self) -> int | None:
        """Return Town hourly forecast humidity."""
        return self.humidity

    def get_precipitation(self) -> float | None:
        """Return Town hourly forecast precipitation."""
        rain = self.get_rain()
        snow = self.get_snow()
        if rain is not None or snow is not None:
            rain = rain or 0.0
            snow = snow or 0.0
            return rain + snow
        return None

    def get_precipitation_probability(self) -> int | None:
        """Return Town hourly forecast precipitation probability."""
        rain_prob = self.get_rain_probability()
        snow_prob = self.get_snow_probability()
        if rain_prob is not None or snow_prob is not None:
            rain_prob = rain_prob or 0
            snow_prob = snow_prob or 0
            return max(rain_prob, snow_prob)
        return None

    def get_rain(self) -> float | None:
        """Return Town hourly forecast rain."""
        return self.rain

    def get_rain_probability(self) -> int | None:
        """Return Town hourly forecast rain probability."""
        return self.rain_probability

    def get_snow(self) -> float | None:
        """Return Town hourly forecast snow."""
        return self.snow

    def get_snow_probability(self) -> int | None:
        """Return Town hourly forecast snow probability."""
        return self.snow_probability

    def get_storm_probability(self) -> int | None:
        """Return Town hourly forecast storm probability."""
        return self.storm_probability

    def get_sunrise(self) -> str:
        """Return Town hourly forecast sunrise."""
        return self.sunrise

    def get_sunset(self) -> str:
        """Return Town hourly forecast sunset."""
        return self.sunset

    def get_temp(self) -> int | None:
        """Return Town hourly forecast temperature."""
        return self.temp

    def get_timestamp(self) -> str:
        """Return Town hourly forecast timestamp."""
        return self._datetime.isoformat()

    def get_wind_direction(self) -> float | None:
        """Return Town hourly forecast wind direction."""
        return self.wind_direction

    def get_wind_speed(self) -> float | None:
        """Return Town hourly forecast wind speed."""
        return self.wind_speed

    def get_wind_speed_max(self) -> float | None:
        """Return Town hourly forecast maximum wind speed."""
        return self.wind_speed_max

    def data(self) -> dict[str, Any]:
        """Return Town hourly forecast data."""
        data: dict[str, Any] = {
            AOD_CONDITION: self.get_condition(),
            AOD_FEEL_TEMP: self.get_feel_temp(),
            AOD_HUMIDITY: self.get_humidity(),
            AOD_PRECIPITATION: self.get_precipitation(),
            AOD_PRECIPITATION_PROBABILITY: self.get_precipitation_probability(),
            AOD_RAIN: self.get_rain(),
            AOD_RAIN_PROBABILITY: self.get_rain_probability(),
            AOD_SNOW: self.get_snow(),
            AOD_SNOW_PROBABILITY: self.get_snow_probability(),
            AOD_STORM_PROBABILITY: self.get_storm_probability(),
            AOD_SUNRISE: self.get_sunrise(),
            AOD_SUNSET: self.get_sunset(),
            AOD_TEMP: self.get_temp(),
            AOD_TIMESTAMP: self.get_timestamp(),
            AOD_WIND_DIRECTION: self.get_wind_direction(),
            AOD_WIND_SPEED: self.get_wind_speed(),
            AOD_WIND_SPEED_MAX: self.get_wind_speed_max(),
        }

        return data

    def parse_interval_value(
        self, values: Any, hour: int, key: str = AEMET_ATTR_VALUE
    ) -> Any:
        """Parse Town hourly forecast interval value from data."""
        period_offset = None
        for value in values:
            if key not in value:
                continue
            period = value[AEMET_ATTR_PERIOD]
            period_start = int(period[0:API_PERIOD_SPLIT])
            if period_offset is None or period_start < period_offset:
                period_offset = period_start
        if period_offset is None:
            period_offset = 0
        for value in values:
            if key not in value:
                continue
            period = value[AEMET_ATTR_PERIOD]
            period_start = int(period[0:API_PERIOD_SPLIT])
            period_end = int(period[API_PERIOD_SPLIT : API_PERIOD_SPLIT * 2])
            period_start -= period_offset
            period_end -= period_offset
            if period_end < period_start:
                period_end = period_end + API_PERIOD_24H
                if hour == 0:
                    hour = hour + API_PERIOD_24H
            if period_start <= hour < period_end:
                return None if not value[key] else value[key]
        return None

    def parse_value(self, values: Any, hour: int, key: str = AEMET_ATTR_VALUE) -> Any:
        """Parse Town hourly forecast value from data."""
        for value in values:
            if key not in value:
                continue
            if int(value[AEMET_ATTR_PERIOD]) == hour:
                return None if not value[key] else value[key]
        return None



"""AEMET OpenData API exceptions."""
from __future__ import annotations


class AemetError(Exception):
    """Base class for AEMET OpenData errors."""


class AuthError(AemetError):
    """Exception raised when API denies access."""


class ApiError(AemetError):
    """Exception raised when data is not provided by API."""


class StationNotFound(AemetError):
    """Exception raised when there are no stations close to provided coordinates."""


class TooManyRequests(AemetError):
    """Exception raised when max API requests are exceeded."""


class TownNotFound(AemetError):
    """Exception raised when there are no towns close to provided coordinates."""



"""AEMET OpenData Helpers."""

from datetime import datetime
from typing import Any
from zoneinfo import ZoneInfo

from .const import (
    AEMET_ATTR_PERIOD,
    AEMET_ATTR_VALUE,
    API_ID_PFX,
    API_PERIOD_24H,
    API_PERIOD_FULL_DAY,
    API_PERIOD_SPLIT,
)

TZ_UTC = ZoneInfo("UTC")


def dict_nested_value(data: dict[str, Any] | None, keys: list[str] | None) -> Any:
    """Get value from dict with nested keys."""
    if keys is None or len(keys) == 0:
        return None
    for key in keys or {}:
        if data is not None:
            data = data.get(key)
    return data


def get_current_datetime(tz: ZoneInfo = TZ_UTC) -> datetime:
    """Return current datetime in UTC."""
    return datetime.now(tz=tz).replace(minute=0, second=0, microsecond=0)


def get_forecast_day_value(
    values: dict[str, Any] | list[Any], key: str = AEMET_ATTR_VALUE
) -> Any:
    """Get day value from forecast."""
    if isinstance(values, list):
        if len(values) > 1:
            for value in values:
                if key not in value:
                    continue
                if value[AEMET_ATTR_PERIOD] == API_PERIOD_FULL_DAY:
                    return value[key]
        else:
            if key in values[0]:
                return values[0][key]
    if isinstance(values, dict):
        if key in values:
            return values[key]
    return None


def get_forecast_hour_value(values: Any, hour: int, key: str = AEMET_ATTR_VALUE) -> Any:
    """Get hour value from forecast."""
    for value in values:
        if key not in value:
            continue
        if int(value[AEMET_ATTR_PERIOD]) == hour:
            return None if not value[key] else value[key]
    return None


def get_forecast_interval_value(
    values: Any, hour: int, key: str = AEMET_ATTR_VALUE
) -> Any:
    """Get hour value from forecast interval."""
    for value in values:
        if key not in value:
            continue
        period_start = int(value[AEMET_ATTR_PERIOD][0:API_PERIOD_SPLIT])
        period_end = int(
            value[AEMET_ATTR_PERIOD][API_PERIOD_SPLIT : API_PERIOD_SPLIT * 2]
        )
        if period_end < period_start:
            period_end = period_end + API_PERIOD_24H
            if hour == 0:
                hour = hour + API_PERIOD_24H
        if period_start <= hour < period_end:
            return None if not value[key] else value[key]
    return None


def split_coordinate(coordinate: str) -> str:
    """Split climatological values station coordinate."""
    coord_deg = coordinate[0:2]
    coord_min = coordinate[2:4]
    coord_sec = coordinate[4:6]
    coord_dir = coordinate[6:7]
    return f"{coord_deg} {coord_min}m {coord_sec}s {coord_dir}"


def parse_api_timestamp(timestamp: str, tz: ZoneInfo = TZ_UTC) -> datetime:
    """Parse AEMET OpenData timestamp into datetime."""
    return datetime.fromisoformat(timestamp).replace(tzinfo=tz)


def parse_station_coordinates(latitude: str, longitude: str) -> str:
    """Parse climatological values station coordinates."""
    return f"{split_coordinate(latitude)} {split_coordinate(longitude)}"


def parse_town_code(town_id: str) -> str:
    """Parse town code from ID if needed."""
    if isinstance(town_id, str) and town_id.startswith(API_ID_PFX):
        return town_id[len(API_ID_PFX) :]
    return town_id


def timezone_from_coords(coords: tuple[float, float]) -> ZoneInfo:
    """Convert coordinates to timezone."""
    if coords[0] < 32 and coords[1] < -11.5:
        return ZoneInfo("Atlantic/Canary")
    return ZoneInfo("Europe/Madrid")



"""AEMET OpenData Town."""

from datetime import datetime
from typing import Any
from zoneinfo import ZoneInfo

from .const import (
    AEMET_ATTR_DATE,
    AEMET_ATTR_DAY,
    AEMET_ATTR_ELABORATED,
    AEMET_ATTR_FORECAST,
    AEMET_ATTR_ID,
    AEMET_ATTR_NAME,
    AEMET_ATTR_TOWN_ALTITUDE,
    AEMET_ATTR_TOWN_LATITUDE_DECIMAL,
    AEMET_ATTR_TOWN_LONGITUDE_DECIMAL,
    AEMET_ATTR_TOWN_RESIDENTS,
    AOD_ALTITUDE,
    AOD_CONDITION,
    AOD_COORDS,
    AOD_DISTANCE,
    AOD_FEEL_TEMP,
    AOD_FORECAST,
    AOD_FORECAST_CURRENT,
    AOD_FORECAST_DAILY,
    AOD_FORECAST_HOURLY,
    AOD_HUMIDITY,
    AOD_ID,
    AOD_NAME,
    AOD_PRECIPITATION,
    AOD_PRECIPITATION_PROBABILITY,
    AOD_RAIN,
    AOD_RAIN_PROBABILITY,
    AOD_RESIDENTS,
    AOD_SNOW,
    AOD_SNOW_PROBABILITY,
    AOD_STORM_PROBABILITY,
    AOD_TEMP,
    AOD_TIMESTAMP,
    AOD_TIMEZONE,
    AOD_UV_INDEX,
    AOD_WIND_DIRECTION,
    AOD_WIND_SPEED,
    AOD_WIND_SPEED_MAX,
    ATTR_DATA,
    ATTR_DISTANCE,
)
from .forecast import DailyForecastValue, HourlyForecastValue
from .helpers import get_current_datetime, parse_api_timestamp, timezone_from_coords


class DailyForecast:
    """AEMET OpenData Town Daily Forecast."""

    _datetime: datetime
    forecast: list[DailyForecastValue]
    zoneinfo: ZoneInfo

    def __init__(self, data: dict[str, Any], zoneinfo: ZoneInfo) -> None:
        """Init AEMET OpenData Town Daily Forecast."""
        self._datetime = parse_api_timestamp(data[AEMET_ATTR_ELABORATED])
        self.forecast: list[DailyForecastValue] = []
        self.zoneinfo = zoneinfo

        cur_dt = get_current_datetime(zoneinfo)
        cur_day = cur_dt.date()

        for day_data in data[AEMET_ATTR_FORECAST][AEMET_ATTR_DAY]:
            day = parse_api_timestamp(day_data[AEMET_ATTR_DATE], zoneinfo)
            if cur_day <= day.date():
                try:
                    self.forecast += [DailyForecastValue(day_data, day)]
                except ValueError:
                    pass

    def get_current_forecast(self) -> DailyForecastValue | None:
        """Return Town current daily forecast."""
        cur_date = get_current_datetime(self.get_timezone()).date()
        for forecast in self.forecast:
            forecast_date = forecast.get_datetime().date()
            if cur_date == forecast_date:
                return forecast
        return None

    def get_timestamp(self) -> str:
        """Return Town daily forecast timestamp."""
        return self._datetime.isoformat()

    def get_timezone(self) -> ZoneInfo:
        """Return Town daily forecast timezone."""
        return self.zoneinfo

    def data(self) -> dict[str, Any]:
        """Return Town daily forecast data."""
        data: dict[str, Any] = {
            AOD_FORECAST: [],
            AOD_TIMESTAMP: self.get_timestamp(),
            AOD_TIMEZONE: self.get_timezone(),
        }

        cur_date = get_current_datetime(self.get_timezone()).date()
        for forecast in self.forecast:
            forecast_date = forecast.get_datetime().date()
            if cur_date <= forecast_date:
                data[AOD_FORECAST] += [forecast.data()]

        cur_forecast = self.get_current_forecast()
        if cur_forecast is not None:
            data[AOD_FORECAST_CURRENT] = cur_forecast.data()

        return data

    def weather(self) -> dict[str, Any]:
        """Return Town daily weather data."""
        weather: dict[str, Any] = {}

        forecast = self.get_current_forecast()
        if forecast is not None:
            weather[AOD_CONDITION] = forecast.get_condition()
            weather[AOD_PRECIPITATION_PROBABILITY] = forecast.get_precipitation_prob()
            weather[AOD_UV_INDEX] = forecast.get_uv_index()
            weather[AOD_WIND_DIRECTION] = forecast.get_wind_direction()
            weather[AOD_WIND_SPEED] = forecast.get_wind_speed()

        return weather


class HourlyForecast:
    """AEMET OpenData Town Hourly Forecast."""

    _datetime: datetime
    forecast: list[HourlyForecastValue]
    zoneinfo: ZoneInfo

    def __init__(self, data: dict[str, Any], zoneinfo: ZoneInfo) -> None:
        """Init AEMET OpenData Town Hourly Forecast."""
        self._datetime = parse_api_timestamp(data[AEMET_ATTR_ELABORATED])
        self.forecast: list[HourlyForecastValue] = []
        self.zoneinfo = zoneinfo

        cur_dt = get_current_datetime(zoneinfo)
        cur_day = cur_dt.date()
        cur_hour = cur_dt.hour

        for day_data in data[AEMET_ATTR_FORECAST][AEMET_ATTR_DAY]:
            day = parse_api_timestamp(day_data[AEMET_ATTR_DATE], zoneinfo)
            day_date = day.date()
            if cur_day <= day_date:
                if cur_day == day_date:
                    start_hour = cur_hour
                else:
                    start_hour = 0

                for hour in range(start_hour, 24):
                    try:
                        cur_forecast = HourlyForecastValue(day_data, day, hour)
                        self.forecast += [cur_forecast]
                    except ValueError:
                        pass

    def get_current_forecast(self) -> HourlyForecastValue | None:
        """Return Town current hourly forecast."""
        cur_dt = get_current_datetime(self.get_timezone())
        for forecast in self.forecast:
            forecast_dt = forecast.get_datetime()
            if cur_dt == forecast_dt:
                return forecast
        return None

    def get_timestamp(self) -> str:
        """Return Town hourly forecast timestamp."""
        return self._datetime.isoformat()

    def get_timezone(self) -> ZoneInfo:
        """Return Town hourly forecast timezone."""
        return self.zoneinfo

    def data(self) -> dict[str, Any]:
        """Return Town hourly forecast data."""
        data: dict[str, Any] = {
            AOD_FORECAST: [],
            AOD_TIMESTAMP: self.get_timestamp(),
            AOD_TIMEZONE: self.get_timezone(),
        }

        cur_dt = get_current_datetime(self.get_timezone())
        for forecast in self.forecast:
            forecast_dt = forecast.get_datetime()
            if cur_dt <= forecast_dt:
                data[AOD_FORECAST] += [forecast.data()]

        cur_forecast = self.get_current_forecast()
        if cur_forecast is not None:
            data[AOD_FORECAST_CURRENT] = cur_forecast.data()

        return data

    def weather(self) -> dict[str, Any]:
        """Return Town hourly weather data."""
        weather: dict[str, Any] = {}

        forecast = self.get_current_forecast()
        if forecast is not None:
            weather[AOD_FEEL_TEMP] = forecast.get_feel_temp()
            weather[AOD_CONDITION] = forecast.get_condition()
            weather[AOD_HUMIDITY] = forecast.get_humidity()
            weather[AOD_PRECIPITATION] = forecast.get_precipitation()
            weather[
                AOD_PRECIPITATION_PROBABILITY
            ] = forecast.get_precipitation_probability()
            weather[AOD_TEMP] = forecast.get_temp()
            weather[AOD_RAIN] = forecast.get_rain()
            weather[AOD_RAIN_PROBABILITY] = forecast.get_rain_probability()
            weather[AOD_SNOW] = forecast.get_snow()
            weather[AOD_SNOW_PROBABILITY] = forecast.get_snow_probability()
            weather[AOD_STORM_PROBABILITY] = forecast.get_storm_probability()
            weather[AOD_WIND_DIRECTION] = forecast.get_wind_direction()
            weather[AOD_WIND_SPEED] = forecast.get_wind_speed()
            weather[AOD_WIND_SPEED_MAX] = forecast.get_wind_speed_max()

        return weather


class Town:
    """AEMET OpenData Town."""

    altitude: int
    coords: tuple[float, float]
    daily: DailyForecast | None
    distance: float
    id: str
    name: str
    residents: int
    zoneinfo: ZoneInfo

    def __init__(self, data: dict[str, Any]) -> None:
        """Init AEMET OpenData Town."""
        self.altitude = int(data[AEMET_ATTR_TOWN_ALTITUDE])
        self.coords = (
            float(data[AEMET_ATTR_TOWN_LATITUDE_DECIMAL]),
            float(data[AEMET_ATTR_TOWN_LONGITUDE_DECIMAL]),
        )
        self.daily: DailyForecast | None = None
        self.distance = float(data[ATTR_DISTANCE])
        self.hourly: HourlyForecast | None = None
        self.id = str(data[AEMET_ATTR_ID])
        self.name = str(data[AEMET_ATTR_NAME])
        self.residents = int(data[AEMET_ATTR_TOWN_RESIDENTS])

        self.zoneinfo = timezone_from_coords(self.coords)

    def get_altitude(self) -> int:
        """Return Town altitude."""
        return self.altitude

    def get_coords(self) -> tuple[float, float]:
        """Return Town coordinates."""
        return self.coords

    def get_distance(self) -> float:
        """Return Town distance from selected coordinates."""
        return round(self.distance, 3)

    def get_id(self) -> str:
        """Return Town ID."""
        return self.id

    def get_residents(self) -> int:
        """Return Town residents."""
        return self.residents

    def get_name(self) -> str:
        """Return Town name."""
        return self.name

    def get_timezone(self) -> ZoneInfo:
        """Return Town zoneinfo."""
        return self.zoneinfo

    def update_daily(self, forecast: dict[str, Any]) -> None:
        """Update Town daily forecast."""
        self.daily = DailyForecast(forecast[ATTR_DATA][0], self.get_timezone())

    def update_hourly(self, forecast: dict[str, Any]) -> None:
        """Update Town hourly forecast."""
        self.hourly = HourlyForecast(forecast[ATTR_DATA][0], self.get_timezone())

    def data(self) -> dict[str, Any]:
        """Return Town data."""
        data: dict[str, Any] = {
            AOD_ALTITUDE: self.get_altitude(),
            AOD_COORDS: self.get_coords(),
            AOD_DISTANCE: self.get_distance(),
            AOD_ID: self.get_id(),
            AOD_NAME: self.get_name(),
            AOD_RESIDENTS: self.get_residents(),
            AOD_TIMEZONE: self.get_timezone(),
        }

        if self.daily is not None:
            data[AOD_FORECAST_DAILY] = self.daily.data()

        if self.hourly is not None:
            data[AOD_FORECAST_HOURLY] = self.hourly.data()

        return data

    def weather_daily(self) -> dict[str, Any]:
        """Return Town daily weather data."""
        if self.daily is not None:
            return self.daily.weather()
        return {}

    def weather_hourly(self) -> dict[str, Any]:
        """Return Town hourly weather data."""
        if self.hourly is not None:
            return self.hourly.weather()
        return {}
